@@ -6,9 +6,26 @@ import string
 import shutil
 import textwrap
 import time
+import signal
+import re
 
 from .Utils import *
 from .Printers import *
+from .EmailService import *
+
+
+def sigint_handler_wrapper(obj: "MatrixTestRunner"):
+    """
+    ``SIGINT`` handler wrapper. Upon SIGINT (Ctrl-C is pressed), output the results already collected to ``stdout`` then abort.
+    """
+    def sigint_handler(signum, frame):
+        print_warning("Process killed by user.")
+        terminal_width, _ = shutil.get_terminal_size()
+        with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.width', terminal_width):  # more options can be specified also
+            print(obj.get_partial_result())
+
+        print_aborted()
+    return sigint_handler
 
 
 class MatrixTestRunner:
@@ -65,7 +82,8 @@ class MatrixTestRunner:
         return fields_matrix == fields_cmd
 
     def __init__(self, cmd: str, matrix: Dict[str, List[str]], parser: Callable[[str], Any] = None, enable_echo: bool = False,
-                 logfile: str = None, timing: bool = False):
+                 logfile: str = None, timing: bool = False, enable_email_notification: bool = False,
+                 email_provider: EmailProvider = None, email_recipient: str = ""):
         """
         Instantiate ``MatrixTestRunner``, checking the user input and initializing options.
 
@@ -87,6 +105,14 @@ class MatrixTestRunner:
 
                 If the result returned from parser function includes a key of "time" (this is how the execution time is recorded,
                 this feature will be disabled and a warning will be displayed.
+        :param enable_email_notification: If ``True``, an email will be sent at the end of :func:`run()`, and the email
+            service provider and the recipient must be given either by the following two arguments or by calling
+            :func:`enable_email_notification()`.
+
+            Please note: if you also set ``send_by_email`` when calling :func:`to_excel()`, you will receive two emails
+            with same content but one of them has attachment.
+        :param email_provider: Instance of email service provider. Refer to :mod:`EmailService`.
+        :param email_recipient: The email address of recipient. Only one recipient is allowed.
         """
 
         colorama.init()  # enable ANSI support on Windows for colored output
@@ -122,6 +148,7 @@ class MatrixTestRunner:
 
         # declare variables for a single run
         self.__last_result = None   # type: Optional[pd.DataFrame]
+        self.__last_record = None   # type: Optional[dict[str, Any]]
         self.__last_repeat = 0
         self.__last_aggregated_columns = []
         # log to file
@@ -139,6 +166,15 @@ class MatrixTestRunner:
         self.__option_echo = enable_echo
         self.__terminal_width, _ = shutil.get_terminal_size()       # this is used by self.__option_echo
 
+        # email
+        # argumetn check will be performed in __send_email()
+        self.__enable_email_notification = enable_email_notification        # type: bool
+        self.__email_provider = email_provider        # type: Optional[EmailProvider]
+        self.__email_recipient = email_recipient      # type: Optional[str]
+
+        # Ctrl-C handler
+        signal.signal(signal.SIGINT, sigint_handler_wrapper(self))
+
     def run(self, repeat: int = 1) -> None:
         """
         Run the test suite and record the results.
@@ -146,6 +182,10 @@ class MatrixTestRunner:
         :param repeat: The number of times the test should be repeated for.
         :return: None. The results will be stored internally. You can use :func:`get_last_result` to get the last result
             as a `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
+
+        You can press ``Ctrl-C`` to interrupt the process and get the partial results that have already collected.
+        Because Python runtime will forward signal to the child process so the running command will be interrupted also
+        and you may see an error from its stdout.
         """
         # log to file
         if self.__log_enabled:
@@ -164,7 +204,7 @@ class MatrixTestRunner:
         results_columns += self.__arg_keys
         # for i in range(repeat):
         #     results_columns.append("attempt" + str(i+1))
-        results = pd.DataFrame(columns=results_columns)
+        self.__last_result = pd.DataFrame(columns=results_columns)
 
         # configure the arg fields and run
         # init
@@ -237,7 +277,9 @@ class MatrixTestRunner:
                         record["attempt" + str(attempt + 1) + "_" + key] = current_result_parsed[key]
                 else:       # single result
                     record["attempt" + str(attempt + 1)] = current_result_parsed
-            results = results.append(record, ignore_index=True)
+                self.__last_record = record
+            self.__last_result = self.__last_result.append(record, ignore_index=True)
+            self.__last_record = None       # reset on the row completion
 
             # get the next or break
             i = self.__nargs - 1
@@ -251,12 +293,17 @@ class MatrixTestRunner:
             if i == -1:  # finished all test cases
                 break
 
-        self.__last_result = results
         print_plain("All done.", self.__log_fd)
+
+        # close log file
         if self.__log_enabled:
             self.__log_fd.close()
             self.__log_fd = None
 
+        # send email
+        if self.__enable_email_notification:
+            self.__send_email()
+
     def get_last_result(self) -> pd.DataFrame:
         """
 
@@ -264,6 +311,15 @@ class MatrixTestRunner:
         """
         return self.__last_result
 
+    def get_partial_result(self) -> pd.DataFrame:
+        """
+
+        :return: Return the internal dataframe storing the final results. The last row may not be complete.
+        """
+        if self.__last_record is not None:
+            self.__last_result = self.__last_result.append(self.__last_record, ignore_index=True)
+        return self.__last_result
+
     def average(self, column: Union[str, List[str]] = None) -> None:
         """
         This function compute the arithmetic mean of selected columns.
@@ -317,14 +373,19 @@ class MatrixTestRunner:
             self.__last_result["avg_"+item] = self.__last_result[columns_in_result].mean(axis=1, numeric_only=True)
             self.__last_aggregated_columns.append("avg_"+item)
 
-    def to_excel(self, path: str, include_agg: bool = True, include_raw: bool = True) -> None:
+    def to_excel(self, path: str, include_agg: bool = True, include_raw: bool = True, send_by_email: bool = False) -> None:
         """
         Export to Excel spreadsheet.
 
         :param path: str. The file path to output.
         :param include_agg: bool. Indicates whether include the aggregated columns or not, such as those generated by average().
         :param include_raw: bool. Indicates whether include the original results. If it is set to False but have no
-                            aggregated results, this argument will be ignored.
+            aggregated results, this argument will be ignored.
+        :param send_by_email: If ``True``, the generated Excel file will be sent via email. The email provider and recipient
+            must be registered at initialization or by calling :func:`enable_email_notification()`.
+
+            Please note: if you also set ``enable_email_notification`` at initialization or by calling :func:`enable_email_notification()`,
+            you will receive two emails with same content but one of them has attachment.
         :return: None
         """
 
@@ -356,6 +417,34 @@ class MatrixTestRunner:
         self.__last_result.to_excel(path, sheet_name="MatrixTest", columns=columns_list, index=False)
         print_ok("Done.")
 
+        if send_by_email:
+            self.__send_email(path)
+
+    def __send_email(self, attachment_path: str = None) -> None:
+        """
+        Wrapper function to send email notification.
+
+        :return:
+        """
+        print_plain("Sending email notification...", self.__log_fd, end="")
+
+        # check provider and recipient is given
+        if self.__email_provider is None or self.__email_recipient is None:
+            print_warning("Email service provider or recipient is not registered at initialization or by calling enable_email_notification().", self.__log_fd)
+            return
+
+        # check email address format
+        if not re.fullmatch("[^@]+@[^@]+\.[^@]+", self.__email_recipient):
+            print_warning("Wrong email address format.", self.__log_fd)
+            return
+
+        # attachment file check will be performed inside the provider because platform-specified requirements.
+        if self.__email_provider.send(self.__matrix, self.__email_recipient, attachment_path):
+            print_ok("Done.", self.__log_fd)
+            print_plain("Please remember to check your junk/spam box.")
+        else:
+            print_warning("Failed.", self.__log_fd)
+
     def enable_echo(self) -> None:
         """
         Enable echo feature, output the ``cmd``'s ``stdout`` in real time.
@@ -374,6 +463,15 @@ class MatrixTestRunner:
         """
         self.__option_echo = False
 
+    def enable_log(self, logfile: str) -> None:
+        """
+        Enable log to file feature. Alias of :func:`change_logfile()`.
+
+        :param logfile: path to the log file
+        :return: None
+        """
+        self.change_logfile(logfile)
+
     def disable_log(self) -> None:
         """
         Disable log to file feature.
@@ -417,3 +515,47 @@ class MatrixTestRunner:
         :return: None
         """
         self.__timing_enabled = False
+
+    def register_email_service(self, provider: EmailProvider, recipient: str) -> None:
+        """
+        Register email service provider and recipient. This function will NOT enable email notification. Please call
+        :func:`enable_email_notification()` to explicitly enable it.
+
+        :param provider: The instance of an email provider class. Refer to :mod:`EmailService`.
+        :param recipient: The email address of recipient. Only one recipient is allowed.
+        :return: None
+        """
+        self.__email_provider = provider
+        self.__email_recipient = recipient
+
+    def deregister_email_service(self) -> None:
+        """
+        Deregister email service provider and recipient, and disable email notification.
+
+        :return: None
+        """
+        self.__email_provider = None
+        self.__email_recipient = None
+        self.__enable_email_notification = False
+
+    def enable_email_notification(self) -> None:
+        """
+        Enable email notification. An email will be sent at the end of :func:`run()`, and the email
+        service provider and the recipient must be given either by the following two arguments or by calling
+        :func:`enable_email_notification()`, otherwise this function has no effect.
+
+        Please note: if you also set ``send_by_email`` when calling :func:`to_excel()`, you will receive two emails
+        with same content but one of them has attachment.
+
+        :return: None
+        """
+        self.__enable_email_notification = True
+
+    def disable_email_notification(self) -> None:
+        """
+        Disable email notification but keep the email service provider and recipient registered, which can be used to send
+        the Excel file by email.
+
+        :return: None
+        """
+        self.__enable_email_notification = False
